/*
 * Gimbal
 * Copyright (C) 2024 moonleay
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package net.moonleay.gimbal.networking

import kotlinx.serialization.ExperimentalSerializationApi
import kotlinx.serialization.cbor.Cbor
import kotlinx.serialization.decodeFromByteArray
import kotlinx.serialization.encodeToByteArray
import net.fabricmc.fabric.api.client.networking.v1.ClientPlayNetworking
import net.fabricmc.fabric.api.networking.v1.PacketByteBufs
import net.minecraft.network.PacketByteBuf
import net.moonleay.gimbal.build.BuildConstants
import net.moonleay.gimbal.client.editor.ClientEditor
import net.moonleay.gimbal.constants.PacketIDs
import net.moonleay.gimbal.editor.state.EditorState
import net.moonleay.gimbal.editor.state.GimbalPolicyType
import net.moonleay.gimbal.editor.state.GimbalServerState

object GimbalClient {

    fun registerPacketHandlers() {
        ClientPlayNetworking.registerGlobalReceiver(PacketIDs.GIMBAL_POLICY_PACKET_ID) { _, _, buf, handler ->
            val state = this.getServerState(buf)
            val isUsable = this.isGimbalUsable(state)
            if (isUsable) {
                val buf = PacketByteBufs.create()
                buf.writeByteArray(Cbor.encodeToByteArray(ClientEditor.getClientState()))
                handler.sendPacket(PacketIDs.UPDATE_EDITOR_STATE_ID, buf)
            }
            ClientEditor.onPolicyReceived(state, false)
        }
    }

    @OptIn(ExperimentalSerializationApi::class)
    private fun getServerState(buf: PacketByteBuf): GimbalServerState {
        val serverState = Cbor.decodeFromByteArray<GimbalServerState>(buf.readByteArray())
        return serverState
    }

    private fun isGimbalUsable(state: GimbalServerState): Boolean {
        return state.policyType == GimbalPolicyType.ALLOWED && state.protocolVersion.toString() == BuildConstants.protocolVersion
    }

    @OptIn(ExperimentalSerializationApi::class)
    private fun createStatePacket(state: EditorState): PacketByteBuf {
        val buf = PacketByteBufs.create()
        buf.writeByteArray(Cbor.encodeToByteArray(state))
        return buf
    }

    fun sendStatePacket(state: EditorState) {
        ClientPlayNetworking.send(PacketIDs.UPDATE_EDITOR_STATE_ID, createStatePacket(state))
    }
}
